"""
Module to provide tests related to the MD029 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest

# pylint: disable=too-many-lines


@pytest.mark.rules
def test_md029_bad_configuration_style():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=$#1",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_list.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md029.style' must be of type 'str'."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_configuration_style_invalide():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=not-matching",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_list.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md029.style' is not valid: Allowable values: ['one', 'ordered', 'zero', 'one_or_ordered']"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/good_one_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_one_three_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/bad_one_one_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_one_three_list.md:3:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 3; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_two_one_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/bad_one_two_one_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_two_one_list.md:3:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 3; Actual: 1; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_two_three_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/good_one_two_three_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_two_three_four_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/bad_two_three_four_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_two_three_four_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 2; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_one_two_three_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/good_zero_one_two_three_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md029/good_zero_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_zero_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 0; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_one_three_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_one_one_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_one_three_list.md:3:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 3; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_two_one_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_one_two_one_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_two_one_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 2; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_two_three_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_two_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_one_two_three_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 2; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_two_three_four_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_two_three_four_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_two_three_four_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 2; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_one_two_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_zero_one_two_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_zero_one_two_three_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 0; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_list_with_config_one():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_zero_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_zero_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 0; Style: 1/1/1] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_one_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 2; Actual: 1; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_one_three_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_one_one_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_one_three_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 2; Actual: 1; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_two_one_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_one_two_one_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_two_one_list.md:3:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 3; Actual: 1; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_two_three_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_two_three_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_two_three_four_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_two_three_four_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_two_three_four_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 2; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_one_two_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_zero_one_two_three_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_list_with_config_ordered():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=ordered",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_zero_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_zero_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 1; Actual: 0; Style: 1/2/3] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_one_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 0; Actual: 1; Style: 0/0/0] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_one_three_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_one_one_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_one_three_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 0; Actual: 1; Style: 0/0/0] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_one_two_one_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_one_two_one_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_one_two_one_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 0; Actual: 1; Style: 0/0/0] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_one_two_three_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_one_two_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_one_two_three_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 0; Actual: 1; Style: 0/0/0] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_bad_two_three_four_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/bad_two_three_four_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/bad_two_three_four_list.md:1:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 0; Actual: 2; Style: 0/0/0] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_one_two_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_zero_one_two_three_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md029/good_zero_one_two_three_list.md:2:1: "
        + "MD029: Ordered list item prefix "
        + "[Expected: 0; Actual: 1; Style: 0/0/0] (ol-prefix)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md029_good_zero_list_with_config_zero():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md029.style=zero",
        "--strict-config",
        "scan",
        "test/resources/rules/md029/good_zero_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#textual-content
"""

import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_textual_content_extra_1():
    """
    Test case extra 1: special characters.
    """

    # Arrange
    source_markdown = """each\bof\athese\x02should\x03be\x04escaped\x05!"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):each\x05\bof\x05\athese\x05\x02should\x05\x03be\x05\x04escaped\x05\x05!:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>each\bof\athese\x02should\x03be\x04escaped\x05!</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_2():
    """
    Test case extra 2: Alert character (\a or replacement character) by itself.
    """

    # Arrange
    source_markdown = """this character \a is me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\a is me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \a is me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_3():
    """
    Test case extra 3: Multiple alert characters (\a or replacement character).
    """

    # Arrange
    source_markdown = """this character \a is me \a or me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\a is me \x05\a or me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \a is me \a or me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_4():
    """
    Test case extra 4: Multiple alert characters and actual replacements
    """

    # Arrange
    source_markdown = """\a&\a&\a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\x05\a\a&\a&amp;\a\x05\a\a&\a&amp;\a\x05\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>\a&amp;\a&amp;\a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_5():
    """
    Test case extra 4: Multiple alert characters and actual replacements in atx
    """

    # Arrange
    source_markdown = """# \a&\a&\a"""
    expected_tokens = [
        "[atx(1,1):1:0:]",
        "[text(1,3):\x05\a\a&\a&amp;\a\x05\a\a&\a&amp;\a\x05\a: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h1>\a&amp;\a&amp;\a</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_6():
    """
    Test case extra 4: Multiple alert characters and actual replacements in setext
    """

    # Arrange
    source_markdown = """\a&\a&\a
---"""
    expected_tokens = [
        "[setext(2,1):-:3::(1,1)]",
        "[text(1,1):\x05\a\a&\a&amp;\a\x05\a\a&\a&amp;\a\x05\a:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>\a&amp;\a&amp;\a</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_7():
    """
    Test case extra 4: Multiple alert characters and actual replacements
    """

    # Arrange
    source_markdown = """`\a<\a>\a`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):\x05\a\a<\a&lt;\a\x05\a\a>\a&gt;\a\x05\a:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>\a&lt;\a&gt;\a</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_8():
    """
    Test case extra 8: Single backspace
    """

    # Arrange
    source_markdown = """this character \b is me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\b is me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \b is me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_9():
    """
    Test case extra 9: Multiple backspace characters
    """

    # Arrange
    source_markdown = """this character \b is me \b or me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\b is me \x05\b or me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \b is me \b or me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_10():
    """
    Test case extra 10: Multiple backspace characters and actual backspaces
    """

    # Arrange
    source_markdown = """\b\\`\b\\`\b"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\x05\b\\\b`\x05\b\\\b`\x05\b:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>\b`\b`\b</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_11():
    """
    Test case extra 11: Multiple backspace characters and actual backspaces in atx
    """

    # Arrange
    source_markdown = """# \b\\`\b\\`\b"""
    expected_tokens = [
        "[atx(1,1):1:0:]",
        "[text(1,3):\x05\b\\\b`\x05\b\\\b`\x05\b: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h1>\b`\b`\b</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_12():
    """
    Test case extra 12: Multiple backspace characters and actual backspaces in setext
    """

    # Arrange
    source_markdown = """\b\\`\b\\`\b
---"""
    expected_tokens = [
        "[setext(2,1):-:3::(1,1)]",
        "[text(1,1):\x05\b\\\b`\x05\b\\\b`\x05\b:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>\b`\b`\b</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_13():
    """
    Test case extra 13: Multiple backslash characters and actual backspaces in link
    """

    # Arrange
    source_markdown = """[foo\\*bar](/bar\\* "ti\\*tle")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/bar*:ti*tle:/bar\\*:ti\\*tle::foo\\*bar:False:":: :]',
        "[text(1,2):foo\\\b*bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/bar*" title="ti*tle">foo*bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_14():
    """
    Test case extra 14: Multiple backspace characters and actual backspaces in almost link
    """

    # Arrange
    source_markdown = """[foo\bbar](/bar\b "ti\btle")"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo\x05\bbar:]",
        "[text(1,9):]:]",
        '[text(1,10):(/bar\x05\b \a"\a&quot;\ati\x05\btle\a"\a&quot;\a):]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo\bbar](/bar\b &quot;ti\btle&quot;)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_15():
    """
    Test case extra 15: Backspace character and actual backspaces in link label
    """

    # Arrange
    source_markdown = """[foo\bbar](/bar "title")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/bar:title::::foo\x05\bbar:False:":: :]',
        "[text(1,2):foo\x05\bbar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/bar" title="title">foo\bbar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_16():
    """
    Test case extra 16: Backspace character and actual backspaces in link uri
    """

    # Arrange
    source_markdown = """[foobar](/b\bar "title")"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foobar:]",
        "[text(1,8):]:]",
        '[text(1,9):(/b\x05\bar \a"\a&quot;\atitle\a"\a&quot;\a):]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foobar](/b\bar &quot;title&quot;)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_17():
    """
    Test case extra 17: Backspace character and actual backspaces in link title
    """

    # Arrange
    source_markdown = """[foobar](/bar "ti\btle")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/bar:ti\btle::::foobar:False:":: :]',
        "[text(1,2):foobar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/bar" title="ti\btle">foobar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_18():
    """
    Test case extra 18: Backspace character and actual backspaces in LRD
    """

    # Arrange
    source_markdown = """before[fo\bo]after

[fo\bo]: /bar "ti\\*\btle"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):before:]",
        "[link(1,7):shortcut:/bar:ti*\btle::::fo\x05\bo:::::]",
        "[text(1,8):fo\x05\bo:]",
        "[end-link::]",
        "[text(1,13):after:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::fo\bo:: :/bar:: :ti*\btle:"ti\\*\btle":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p>before<a href="/bar" title="ti*\btle">fo\bo</a>after</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_19():
    """
    Test case extra 19: Backspace character and actual backspaces in fenced block info
    """

    # Arrange
    source_markdown = """``` foo\b\\+\bbar
foo
```"""
    expected_tokens = [
        "[fcode-block(1,1):`:3:foo\b+\bbar:foo\b\\+\bbar:::: ]",
        "[text(2,1):foo:]",
        "[end-fcode-block::3:False]",
    ]
    expected_gfm = """<pre><code class="language-foo\b+\bbar">foo\n</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_20():
    """
    Test case extra 20: Whitespace split character
    NOTE: Since this is only appears not-naturally in whitespace, cannot come
          up with embedded test case.
    """

    # Arrange
    source_markdown = """this character \x02 is me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\x02 is me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \x02 is me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_21():
    """
    Test case extra 21: NOOP character
    NOTE: Since this is only appears not-naturally in whitespace, cannot come
          up with embedded test case.
    """

    # Arrange
    source_markdown = """this character \x03 is me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\x03 is me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \x03 is me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_22():
    """
    Test case extra 22: Blech character
    NOTE: Since this is only appears not-naturally in whitespace, cannot come
          up with embedded test case.
    """

    # Arrange
    source_markdown = """this character \x04 is me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\x04 is me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \x04 is me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_23():
    """
    Test case extra 23: Escape character
    """

    # Arrange
    source_markdown = """this character \x05 is me"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this character \x05\x05 is me:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>this character \x05 is me</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_extra_24():
    """
    Test case extra 24: all special characters, one after the other
    """

    # Arrange
    source_markdown = """\b\a\x02\x03\x04\x05"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\x05\b\x05\a\x05\x02\x05\x03\x05\x04\x05\x05:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>\b\a\x02\x03\x04\x05</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Processing to coalesce a text tokens with a list of tokens.
"""
import logging

from pymarkdown.inline_markdown_token import TextMarkdownToken
from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class CoalesceProcessor:
    """
    Handle the text coalesce processing of the token stream.
    """

    @staticmethod
    def coalesce_text_blocks(first_pass_results):
        """
        Take a pass and combine any two adjacent text blocks into one.
        """
        coalesced_list = []
        coalesced_list.extend(first_pass_results[0:1])
        for coalesce_index in range(1, len(first_pass_results)):
            did_process = False
            POGGER.debug(
                "coalesce_text_blocks>>>>$<<",
                first_pass_results[coalesce_index],
            )
            if coalesced_list[-1].is_text:
                POGGER.debug(">>coalesce_text_blocks>>>>$<<", coalesced_list[-1])
                if first_pass_results[coalesce_index].is_text or (
                    first_pass_results[coalesce_index].is_blank_line
                    and coalesced_list[-2].is_code_block
                ):

                    POGGER.debug("text-text>>$<<", coalesced_list[-2])
                    if coalesced_list[-2].is_indented_code_block:
                        remove_leading_spaces = len(
                            coalesced_list[-2].extracted_whitespace
                        )
                    elif (
                        coalesced_list[-2].is_paragraph
                        or coalesced_list[-2].is_setext_heading
                    ):
                        remove_leading_spaces = -1
                    else:
                        remove_leading_spaces = 0

                    POGGER.debug("remove_leading_spaces>>$", remove_leading_spaces)
                    POGGER.debug("combine1>>$", coalesced_list[-1])
                    POGGER.debug("combine2>>$", first_pass_results[coalesce_index])
                    indented_whitespace = coalesced_list[-1].combine(
                        first_pass_results[coalesce_index], remove_leading_spaces
                    )
                    POGGER.debug("combined>>$", coalesced_list[-1])
                    POGGER.debug("indented_whitespace>>$<<", indented_whitespace)
                    if coalesced_list[-2].is_indented_code_block:
                        coalesced_list[-2].add_indented_whitespace(indented_whitespace)
                    did_process = True
            elif (
                first_pass_results[coalesce_index].is_blank_line
                and coalesced_list[-1].is_code_block
            ):
                POGGER.debug("was>>$", first_pass_results[coalesce_index])
                replacement_token = TextMarkdownToken(
                    "",
                    first_pass_results[coalesce_index].extracted_whitespace,
                    line_number=first_pass_results[coalesce_index].line_number,
                    column_number=first_pass_results[coalesce_index].column_number,
                )
                POGGER.debug("now>>$", replacement_token)
                coalesced_list.append(replacement_token)
                did_process = True
            if not did_process:
                coalesced_list.append(first_pass_results[coalesce_index])

        for coalesce_index in range(1, len(coalesced_list)):
            if coalesced_list[coalesce_index].is_text and (
                coalesced_list[coalesce_index - 1].is_paragraph
                or coalesced_list[coalesce_index - 1].is_setext_heading
            ):
                POGGER.debug("full_paragraph_text>$<", coalesced_list[coalesce_index])
                POGGER.debug(
                    "full_paragraph_text>$<",
                    coalesced_list[coalesce_index].token_text,
                )
                removed_ws = coalesced_list[coalesce_index].remove_final_whitespace()
                POGGER.debug(
                    "full_paragraph_text>$<",
                    coalesced_list[coalesce_index].token_text,
                )
                POGGER.debug(
                    "full_paragraph_text>$>", coalesced_list[coalesce_index - 1]
                )
                coalesced_list[coalesce_index - 1].set_final_whitespace(removed_ws)
                POGGER.debug(
                    "full_paragraph_text>$>", coalesced_list[coalesce_index - 1]
                )

        return coalesced_list


# pylint: enable=too-few-public-methods



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_380():
    """
    Test case 380:  Rule 4:  This is not emphasis, because the closing _ is preceded by whitespace:
    """

    # Arrange
    source_markdown = """_foo bar _"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2):foo bar :]",
        "[text(1,10):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_foo bar _</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_381():
    """
    Test case 381:  This is not emphasis, because the second _ is preceded by punctuation and followed by an alphanumeric:
    """

    # Arrange
    source_markdown = """_(_foo)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2):(:]",
        "[text(1,3):_:]",
        "[text(1,4):foo):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_(_foo)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_382():
    """
    Test case 382:  This is emphasis within emphasis:
    """

    # Arrange
    source_markdown = """_(_foo_)_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):(:]",
        "[emphasis(1,3):1:_]",
        "[text(1,4):foo:]",
        "[end-emphasis(1,7)::]",
        "[text(1,8):):]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>(<em>foo</em>)</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_383():
    """
    Test case 383:  (part 1) Intraword emphasis is disallowed for _:
    """

    # Arrange
    source_markdown = """_foo_bar"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2):foo:]",
        "[text(1,5):_:]",
        "[text(1,6):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_foo_bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_384():
    """
    Test case 384:  (part 2) Intraword emphasis is disallowed for _:
    """

    # Arrange
    source_markdown = """_пристаням_стремятся"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2):пристаням:]",
        "[text(1,11):_:]",
        "[text(1,12):стремятся:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_пристаням_стремятся</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_385():
    """
    Test case 385:  (part 3) Intraword emphasis is disallowed for _:
    """

    # Arrange
    source_markdown = """_foo_bar_baz_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):foo:]",
        "[text(1,5):_:]",
        "[text(1,6):bar:]",
        "[text(1,9):_:]",
        "[text(1,10):baz:]",
        "[end-emphasis(1,13)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo_bar_baz</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_386():
    """
    Test case 386:  This is emphasis, even though the closing delimiter is both left- and right-flanking, because it is followed by punctuation:
    """

    # Arrange
    source_markdown = """_(bar)_."""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):(bar):]",
        "[end-emphasis(1,7)::]",
        "[text(1,8):.:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>(bar)</em>.</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to implement a plugin that looks for trailing spaces in the files.
"""
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.plugin_manager import Plugin, PluginDetails


# pylint: disable=too-many-instance-attributes
class RuleMd009(Plugin):
    """
    Class to implement a plugin that looks for trailing spaces in the files.
    """

    def __init__(self):
        super().__init__()
        self.__leaf_tokens = None
        self.__leaf_owner_tokens = None
        self.__line_index = None
        self.__leaf_token_index = None
        self.__inline_token_index = None
        self.__container_token_stack = None
        self.__break_spaces = None
        self.__strict_mode = None
        self.__list_item_empty_lines_mode = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-trailing-spaces",
            plugin_id="MD009",
            plugin_enabled_by_default=True,
            plugin_description="Trailing spaces",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md009.md",
            plugin_configuration="br_spaces,list_item_empty_lines,strict",
        )

    @classmethod
    def __validate_break_spaces(cls, found_value):
        if found_value < 0:
            raise ValueError("Allowable values are greater than or equal to 0.")

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """
        self.__break_spaces = self.plugin_configuration.get_integer_property(
            "br_spaces",
            default_value=2,
            valid_value_fn=self.__validate_break_spaces,
        )
        if self.__break_spaces < 2:
            self.__break_spaces = 0
        self.__strict_mode = self.plugin_configuration.get_boolean_property(
            "strict",
            default_value=False,
        )
        self.__list_item_empty_lines_mode = (
            self.plugin_configuration.get_boolean_property(
                "list_item_empty_lines",
                default_value=False,
            )
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__leaf_tokens = []
        self.__leaf_owner_tokens = []
        self.__line_index = 1
        self.__leaf_token_index = 0
        self.__inline_token_index = 0
        self.__container_token_stack = []

    def next_line(self, context, line):
        """
        Event that a new line is being processed.
        """
        if (
            self.__leaf_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__leaf_token_index + 1].line_number
            and self.__leaf_tokens[self.__leaf_token_index + 1].is_leaf
        ):
            self.__leaf_token_index = self.__inline_token_index + 1
            self.__inline_token_index = self.__leaf_token_index
        if (
            self.__inline_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__inline_token_index + 1].line_number
        ):
            self.__inline_token_index += 1

        if (
            not self.__leaf_tokens[self.__leaf_token_index].is_code_block
            and line
            and line[-1] == " "
        ):

            (
                first_non_whitespace_index,
                extracted_whitespace,
            ) = ParserHelper.extract_whitespace_from_end(line)
            extracted_whitespace_length = len(extracted_whitespace)

            is_list_empty_line = (
                self.__list_item_empty_lines_mode
                and self.__leaf_owner_tokens[self.__leaf_token_index]
                and self.__leaf_owner_tokens[self.__leaf_token_index].is_list_start
                and first_non_whitespace_index == 0
            )

            if extracted_whitespace_length != self.__break_spaces or (
                self.__strict_mode and not is_list_empty_line
            ):

                if self.__strict_mode or self.__break_spaces < 2:
                    extra_error_information = "0"
                else:
                    extra_error_information = f"0 or {self.__break_spaces}"
                extra_error_information = f"Expected: {extra_error_information}; Actual: {extracted_whitespace_length}"
                self.report_next_line_error(
                    context,
                    first_non_whitespace_index + 1,
                    extra_error_information=extra_error_information,
                )

        self.__line_index += 1

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        _ = context

        if token.is_block_quote_start or token.is_list_start:
            self.__container_token_stack.append(token)
        elif token.is_block_quote_end or token.is_list_end:
            del self.__container_token_stack[-1]
        elif token.is_blank_line or token.is_leaf or token.is_inline_hard_break:
            self.__leaf_tokens.append(token)
            if self.__container_token_stack:
                self.__leaf_owner_tokens.append(self.__container_token_stack[-1])
            else:
                self.__leaf_owner_tokens.append(None)


# pylint: enable=too-many-instance-attributes



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_366():
    """
    Test case 366:  Rule 2:
    """

    # Arrange
    source_markdown = """_foo bar_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):foo bar:]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_367():
    """
    Test case 367:  This is not emphasis, because the opening _ is followed by whitespace:
    """

    # Arrange
    source_markdown = """_ foo bar_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2): foo bar:]",
        "[text(1,10):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_ foo bar_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_368():
    """
    Test case 368:  This is not emphasis, because the opening _ is preceded by an alphanumeric and followed by punctuation:
    """

    # Arrange
    source_markdown = """a_"foo"_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[text(1,2):_:]",
        '[text(1,3):\a"\a&quot;\afoo\a"\a&quot;\a:]',
        "[text(1,8):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a_&quot;foo&quot;_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_369():
    """
    Test case 369:  (part 1) Emphasis with _ is not allowed inside words:
    """

    # Arrange
    source_markdown = """foo_bar_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo:]",
        "[text(1,4):_:]",
        "[text(1,5):bar:]",
        "[text(1,8):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo_bar_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_370():
    """
    Test case 370:  (part 2) Emphasis with _ is not allowed inside words:
    """

    # Arrange
    source_markdown = """5_6_78"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):5:]",
        "[text(1,2):_:]",
        "[text(1,3):6:]",
        "[text(1,4):_:]",
        "[text(1,5):78:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>5_6_78</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_371():
    """
    Test case 371:  (part 3) Emphasis with _ is not allowed inside words:
    """

    # Arrange
    source_markdown = """пристаням_стремятся_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):пристаням:]",
        "[text(1,10):_:]",
        "[text(1,11):стремятся:]",
        "[text(1,20):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>пристаням_стремятся_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_372():
    """
    Test case 372:  Here _ does not generate emphasis, because the first delimiter run is right-flanking and the second left-flanking:
    """

    # Arrange
    source_markdown = """aa_"bb"_cc"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):aa:]",
        "[text(1,3):_:]",
        '[text(1,4):\a"\a&quot;\abb\a"\a&quot;\a:]',
        "[text(1,8):_:]",
        "[text(1,9):cc:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>aa_&quot;bb&quot;_cc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_373():
    """
    Test case 373:  This is emphasis, even though the opening delimiter is both left- and right-flanking, because it is preceded by punctuation:
    """

    # Arrange
    source_markdown = """foo-_(bar)_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo-:]",
        "[emphasis(1,5):1:_]",
        "[text(1,6):(bar):]",
        "[end-emphasis(1,11)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo-<em>(bar)</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide for a debug extension.
"""

from pymarkdown.extension_impl import ExtensionDetails


# pylint: disable=too-few-public-methods
class DebugExtension:
    """
    Extension to provide for a debug extension.
    """

    @classmethod
    def get_details(cls):
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id="debug-extension",
            extension_name="Debug Extension",
            extension_description="Allows testing through debug.",
            extension_enabled_by_default=False,
            extension_version="0.0.0",
            extension_interface_version=1,
            extension_url=None,
            extension_configuration=None,
        )

    @classmethod
    def apply_configuration(cls, extension_specific_facade):
        """
        Apply any configuration required by the extension.
        """
        debug_mode = extension_specific_facade.get_integer_property(
            "debug_mode", default_value=0
        )
        if debug_mode == 1:
            raise Exception("blah")
        if debug_mode == 2:
            raise ValueError("blah")


# pylint: enable=too-few-public-methods



"""
Tests for the optional front-matter processing
"""
import pytest

from .utils import act_and_assert

config_map = {"extensions": {"front-matter": {"enabled": True}}}


@pytest.mark.gfm
def test_front_matter_01():
    """
    Any whitespace before the three - characters causes it not to fire.
    fill in layer - test_thematic_breaks_020
    """

    # Arrange
    source_markdown = """ ---
Title: my document
---
"""
    expected_tokens = [
        "[tbreak(1,2):-: :---]",
        "[setext(3,1):-:3::(2,1)]",
        "[text(2,1):Title: my document:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<hr />
<h2>Title: my document</h2>"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_02():
    """
    The starting character must be the '-' character, not the other two.
    """

    # Arrange
    source_markdown = """***
Title: my document
***
"""
    expected_tokens = [
        "[tbreak(1,1):*::***]",
        "[para(2,1):]",
        "[text(2,1):Title: my document:]",
        "[end-para:::False]",
        "[tbreak(3,1):*::***]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<hr />
<p>Title: my document</p>
<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_03():
    """
    Everything between the start and end is parsed, but not as part of HTML output.
    """

    # Arrange
    source_markdown = """---
Title: my document
---
---
"""
    expected_tokens = [
        "[front-matter(1,1):---:---:['Title: my document']:{'title': 'my document'}]",
        "[tbreak(4,1):-::---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_04():
    """
    There must be an opening and closing boundary for it to be eligible as front matter.
    """

    # Arrange
    source_markdown = """---
"""
    expected_tokens = ["[tbreak(1,1):-::---]", "[BLANK(2,1):]"]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown,
        expected_gfm,
        expected_tokens,
        config_map=config_map,
    )


@pytest.mark.gfm
def test_front_matter_05():
    """
    There must be an opening and closing boundary for it to be eligible as front matter.
    """

    # Arrange
    source_markdown = """---
test:
"""
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[para(2,1):]",
        "[text(2,1):test::]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<hr />
<p>test:</p>"""

    # Act & Assert
    act_and_assert(
        source_markdown,
        expected_gfm,
        expected_tokens,
        config_map=config_map,
    )


@pytest.mark.gfm
def test_front_matter_06():
    """
    There must be an opening and closing boundary for it to be eligible as front matter.
    Even if there is just a field name and no value.
    """

    # Arrange
    source_markdown = """---
test:
---
"""
    expected_tokens = [
        "[front-matter(1,1):---:---:['test:']:{'test': ''}]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_07():
    """
    There must be an opening and closing boundary for it to be eligible as front matter.
    Containing a single line field name and value is normal.
    """

    # Arrange
    source_markdown = """---
test: abc
---
"""
    expected_tokens = [
        "[front-matter(1,1):---:---:['test: abc']:{'test': 'abc'}]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_08():
    """
    There must be an opening and closing boundary for it to be eligible as front matter.
    In normal mode, a multiline field value is indicated by a second line that is indented
    by at least 4 characters.
    """

    # Arrange
    source_markdown = """---
test: abc
    def
---
"""
    expected_tokens = [
        "[front-matter(1,1):---:---:['test: abc', '    def']:{'test': 'abc\\ndef'}]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_09():
    """
    If a field name does not start a new line or there isn't 4+ spaces at the start,
    the entire header is abandoned.
    """

    # Arrange
    source_markdown = """---
test: abc
   def
---
"""
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[setext(4,1):-:3::(2,1)]",
        "[text(2,1):test: abc\ndef::\n   \x02]",
        "[end-setext::]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<hr />
<h2>test: abc
def</h2>"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_10():
    """
    A field name can be indented as many as 3 characters, as long as it ends with a ':'.
    """

    # Arrange
    source_markdown = """---
test: abc
   def:
---
"""
    expected_tokens = [
        "[front-matter(1,1):---:---:['test: abc', '   def:']:{'test': 'abc', 'def': ''}]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_11():
    """
    A front matter element must contain at least one field name.
    """

    # Arrange
    source_markdown = """---
---"""
    expected_tokens = ["[tbreak(1,1):-::---]", "[tbreak(2,1):-::---]"]
    expected_gfm = """<hr />
<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_12():
    """
    A continuation without a field to associate it with is bad.
    """

    # Arrange
    source_markdown = """---
\a\a\a\acontinuation
---""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[icode-block(2,5):    :]",
        "[text(2,5):continuation:]",
        "[end-icode-block:::False]",
        "[tbreak(3,1):-::---]",
    ]
    expected_gfm = """<hr />
<pre><code>continuation
</code></pre>
<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_13():
    """
    If a blank line is encountered before the end marker, the entire header is
    thrown out.
    """

    # Arrange
    source_markdown = """---

Title: my document
---
---
"""
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[BLANK(2,1):]",
        "[setext(4,1):-:3::(3,1)]",
        "[text(3,1):Title: my document:]",
        "[end-setext::]",
        "[tbreak(5,1):-::---]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<hr />
<h2>Title: my document</h2>
<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_14():
    """
    Any whitespace after the three - characters in the start boundary is acceptable.
    """

    # Arrange
    source_markdown = """---\a\a
Title: my document
---
""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[front-matter(1,1):---  :---:['Title: my document']:{'title': 'my document'}]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_15():
    """
    Any whitespace after the three - characters in the end boundary is acceptable.
    """

    # Arrange
    source_markdown = """---
Title: my document
---\a\a
""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[front-matter(1,1):---:---  :['Title: my document']:{'title': 'my document'}]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_16():
    """
    More than three - characters in the boundary is not acceptable.
    """

    # Arrange
    source_markdown = """----
Title: my document
----
"""
    expected_tokens = [
        "[tbreak(1,1):-::----]",
        "[setext(3,1):-:4::(2,1)]",
        "[text(2,1):Title: my document:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<hr />\n<h2>Title: my document</h2>"""

    # Act & Assert
    act_and_assert(
        source_markdown,
        expected_gfm,
        expected_tokens,
        config_map=config_map,
    )


@pytest.mark.gfm
def test_front_matter_17():
    """
    This is an extension of test_front_matter_13. If a blank line is encountered
    before the end marker, but after a field name, the entire header is still thrown out.
    """

    # Arrange
    source_markdown = """---
Title: my document

---
---
"""
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[para(2,1):]",
        "[text(2,1):Title: my document:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[tbreak(4,1):-::---]",
        "[tbreak(5,1):-::---]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<hr />
<p>Title: my document</p>
<hr />
<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_18():
    """
    This is an extension of test_front_matter_13/17. If a blank line is encountered
    before the end marker, but after a field name and indented by at least 4 spaces,
    the front matter is still valid.
    """

    # Arrange
    source_markdown = """---
Title: my document
/a/a/a/a
---
---
""".replace(
        "/a", " "
    )
    expected_tokens = [
        "[front-matter(1,1):---:---:['Title: my document', '    ']:{'title': 'my document\\n'}]",
        "[tbreak(5,1):-::---]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )


@pytest.mark.gfm
def test_front_matter_19():
    """
    This is an extension of test_front_matter_18. If a blank line is encountered
    before the end marker, but before a field name and indented by at least 4 spaces,
    the front matter is no longer valid.
    """

    # Arrange
    source_markdown = """---
/a/a/a/a
Title: my document
---
---
""".replace(
        "/a", " "
    )
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[BLANK(2,1):    ]",
        "[setext(4,1):-:3::(3,1)]",
        "[text(3,1):Title: my document:]",
        "[end-setext::]",
        "[tbreak(5,1):-::---]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<hr />
<h2>Title: my document</h2>
<hr />"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, config_map=config_map
    )



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_t():
    """
    Test case:  Block Quote with text, newline, block quote, text
    """

    # Arrange
    source_markdown = """> uvw
> xyz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):\n]",
        "[text(1,3):uvw\nxyz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw
xyz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_i2_t():
    """
    Test case:  Block Quote with text, newline, indent of 2, text
    """

    # Arrange
    source_markdown = """> uvw
  xyz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):\n  ]",
        "[text(1,3):uvw\nxyz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw
xyz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i2_bq_t():
    """
    Test case:  unordered list, text, newline, indent of 2, block Quote, text, newline,
                indent of 2, block quote, text
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  > xyz
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > ]",
        "[para(2,5):\n]",
        "[text(2,5):uvw\nxyz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw
xyz</p>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i4_t():
    """
    Test case:  unordered list, text, newline, indent of 2, block Quote, text, newline,
                indent of 4, text
    """

    # Arrange
    source_markdown = """* abc
  > uvw
    xyz
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  ]",
        "[para(2,5):\n  ]",
        "[text(2,5):uvw\nxyz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw
xyz</p>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_ha():
    """
    Test case:  Block Quote with text, newline, block quote, atx heading
    """

    # Arrange
    source_markdown = """> uvw
> # head"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::False]",
        "[atx(2,3):1:0:]",
        "[text(2,5):head: ]",
        "[end-atx::]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
<h1>head</h1>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_i2_ha():
    """
    Test case:  Block Quote with text, newline, proper indent for lazy, atx heading
    """

    # Arrange
    source_markdown = """> uvw
  # head"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[atx(2,3):1:0:  ]",
        "[text(2,5):head: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
</blockquote>
<h1>head</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i2_bq_ha():
    """
    Test case:  Unordered list, text, newline, Indent of 2, Block Quote with text,
    newline, ident of 2, block quote, atx heading
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  > # head
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::False]",
        "[atx(3,5):1:0:]",
        "[text(3,7):head: ]",
        "[end-atx::]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
<h1>head</h1>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_i2_bq_t_nl_i4_ha():
    """
    Test case:  Unordered list, text, newline, Indent of 2, Block Quote with text,
    newline, ident of 4, atx heading
    """

    # Arrange
    source_markdown = """* abc
  > uvw
    # head
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[atx(3,5):1:0:  ]",
        "[text(3,7):head: ]",
        "[end-atx::]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
</blockquote>
<h1>head</h1>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_tb():
    """
    Test case:  Block quote, text, newline, block quote, thematic break
    """

    # Arrange
    source_markdown = """> uvw
> ---"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[setext(2,3):-:3::(1,3)]",
        "[text(1,3):uvw:]",
        "[end-setext::]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h2>uvw</h2>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_i2_tb():
    """
    Test case:  Block quote, text, newline, indent of 2, thematic break
    """

    # Arrange
    source_markdown = """> uvw
  ---"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[tbreak(2,3):-:  :---]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
</blockquote>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i2_bq_tb():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 2, block quote, thematic break
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  > ---
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > ]",
        "[setext(3,5):-:3::(2,5)]",
        "[text(2,5):uvw:]",
        "[end-setext::]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<h2>uvw</h2>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i4_tb():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 4, thematic break
    """

    # Arrange
    source_markdown = """* abc
  > uvw
    ---
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[tbreak(3,5):-:  :---]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
</blockquote>
<hr />
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_hb():
    """
    Test case:  Block quote, text, newline, block quote, html block
    """

    # Arrange
    source_markdown = """> uvw
> <!-- comment -->"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::False]",
        "[html-block(2,3)]",
        "[text(2,3):<!-- comment -->:]",
        "[end-html-block:::False]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
<!-- comment -->
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_i2_hb():
    """
    Test case:  Block quote, text, newline, indent of 2, html block
    """

    # Arrange
    source_markdown = """> uvw
  <!-- comment -->"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[html-block(2,1)]",
        "[text(2,3):<!-- comment -->:  ]",
        "[end-html-block:::False]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
</blockquote>
  <!-- comment -->"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i2_bq_hb():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 2, block quote, html block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  > <!-- comment -->
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::False]",
        "[html-block(3,5)]",
        "[text(3,5):<!-- comment -->:]",
        "[end-html-block:::False]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
<!-- comment -->
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i4_hb():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 4, html block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
    <!-- comment -->
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[html-block(3,3)]",
        "[text(3,5):<!-- comment -->:  ]",
        "[end-html-block:::False]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
</blockquote>
  <!-- comment -->
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_fb():
    """
    Test case:  Block quote, text, newline, block quote, fenced block
    """

    # Arrange
    source_markdown = """> uvw
> ```
> def
> ```"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::False]",
        "[fcode-block(2,3):`:3::::::]",
        "[text(3,3):def:]",
        "[end-fcode-block::3:False]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
<pre><code>def
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_fb_nl_with_bq():
    """
    Test case:  Block quote, text, newline, block quote, fenced block with newlines
                prefaced by block quotes
    """

    # Arrange
    source_markdown = """> uvw
> ```
>
> def
>
> ```"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::False]",
        "[fcode-block(2,3):`:3::::::]",
        "[text(3,3):\ndef\n:]",
        "[end-fcode-block::3:False]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
<pre><code>
def

</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_fb_nl_without_bq():
    """
    Test case:  Block quote, text, newline, block quote, fenced block with newlines
                not prefaced by block quotes
    """

    # Arrange
    source_markdown = """> uvw
> ```

> def

> ```"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::False]",
        "[fcode-block(2,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(3,1):]",
        "[block-quote(4,1)::> \n]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(5,1):]",
        "[block-quote(6,1)::> ]",
        "[fcode-block(6,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
<pre><code></code></pre>
</blockquote>
<blockquote>
<p>def</p>
</blockquote>
<blockquote>
<pre><code></code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_i2_fb():
    """
    Test case:  Block quote, text, newline, indent of 2, fenced block
    """

    # Arrange
    source_markdown = """> uvw
  ```
  def
  ```"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(2,3):`:3:::::  :]",
        "[text(3,3):def:\a  \a\x03\a]",
        "[end-fcode-block:  :3:False]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
</blockquote>
<pre><code>def
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i2_bq_fb():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 2, block quote, fenced block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  > ```
  > def
  > ```
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > \n  > \n  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::False]",
        "[fcode-block(3,5):`:3::::::]",
        "[text(4,5):def:]",
        "[end-fcode-block::3:False]",
        "[end-block-quote:::True]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
<pre><code>def
</code></pre>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i4_fb():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 4, fenced block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
    ```
    def
    ```
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(3,5):`:3:::::  :]",
        "[text(4,3):def:\a  \a\x03\a]",
        "[end-fcode-block:  :3:False]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
</blockquote>
<pre><code>def
</code></pre>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_ib():
    """
    Test case:  Block quote, text, newline, block quote, indented block
    """

    # Arrange
    source_markdown = """> uvw
>     def"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):\n    ]",
        "[text(1,3):uvw\ndef::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw
def</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_i4_t_nl_bq_i4_t():
    """
    Test case:  Block quote indent of 4 text newline block quote indent of 4 text
    """

    # Arrange
    source_markdown = """>     foo
>     bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[icode-block(1,7):    :\n    ]",
        "[text(1,7):foo\nbar:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>foo
bar
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_i4_t_nl_bq_i3_t():
    """
    Test case:  Block quote indent of 4 text newline block quote indent of 3 text
    """

    # Arrange
    source_markdown = """>     foo
>    bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[icode-block(1,7):    :]",
        "[text(1,7):foo:]",
        "[end-icode-block:::False]",
        "[para(2,6):   ]",
        "[text(2,6):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>foo
</code></pre>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_i4_t_nl_nl_bq_i4_t():
    """
    Test case:  Block quote indent of 4 text newline newline block quote indent of 4 text
    """

    # Arrange
    source_markdown = """>     foo

>     bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[icode-block(1,7):    :]",
        "[text(1,7):foo:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[icode-block(3,7):    :]",
        "[text(3,7):bar:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>foo
</code></pre>
</blockquote>
<blockquote>
<pre><code>bar
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_i4_t_nl_bq_nl_bq_i4_t():
    """
    Test case:  Block quote indent of 4 text newline block quote newline block quote indent of 4 text
    """

    # Arrange
    source_markdown = """>     foo

>     bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[icode-block(1,7):    :]",
        "[text(1,7):foo:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[icode-block(3,7):    :]",
        "[text(3,7):bar:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>foo
</code></pre>
</blockquote>
<blockquote>
<pre><code>bar
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_bq_nl_bq_ib():
    """
    Test case:  Block quote, text, newline, block quote, newline block quote, indented block
    """

    # Arrange
    source_markdown = """> uvw
>
>     def"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        "[icode-block(3,7):    :]",
        "[text(3,7):def:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
<pre><code>def
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_i6_ib():
    """
    Test case:  Block quote, text, newline, indent of 6, indented block
    """

    # Arrange
    source_markdown = """> uvw
      def"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):\n      ]",
        "[text(1,3):uvw\ndef::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw
def</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_bq_t_nl_nl_nl_i6_ib():
    """
    Test case:  Block quote, text, newline, newline, indent of 6, indented block
    """

    # Arrange
    source_markdown = """> uvw

      def"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[icode-block(3,5):    :]",
        "[text(3,5):def:  ]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<blockquote>
<p>uvw</p>
</blockquote>
<pre><code>  def
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i2_bq_ib():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 2, block quote, indented block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  >     def
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > ]",
        "[para(2,5):\n    ]",
        "[text(2,5):uvw\ndef::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw
def</p>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_bq_nl_i2_bq_ib():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                block quote, newline, indent of 2, block quote, indented block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
  >
  >     def
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  >\n  > ]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::True]",
        "[BLANK(3,4):]",
        "[icode-block(4,9):    :]",
        "[text(4,9):def:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw</p>
<pre><code>def
</code></pre>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_i6_ib():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                indent of 6, indented block
    """

    # Arrange
    source_markdown = """* abc
  > uvw
      def
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  ]",
        "[para(2,5):\n    ]",
        "[text(2,5):uvw\ndef::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
<blockquote>
<p>uvw
def</p>
</blockquote>
</li>
<li>def</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_n_ul_t_nl_i2_bq_t_nl_nl_i6_ib():
    """
    Test case:  Unordered list, text, newline, ident of 2, block quote, text, newline,
                newline, indent of 6, indented block
    """

    # Arrange
    source_markdown = """* abc
  > uvw

      def
* def"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n]",
        "[para(2,5):]",
        "[text(2,5):uvw:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(3,1):]",
        "[icode-block(4,7):    :]",
        "[text(4,7):def:]",
        "[end-icode-block:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):def:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>abc</p>
<blockquote>
<p>uvw</p>
</blockquote>
<pre><code>def
</code></pre>
</li>
<li>
<p>def</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


# setext?



"""
https://github.github.com/gfm/#inlines
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_inlines_307():
    """
    Test case 307:  Thus, for example, in
    """

    # Arrange
    source_markdown = """`hi`lo`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):hi:`::]",
        "[text(1,5):lo`:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>hi</code>lo`</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide tests related to the MD037 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest


@pytest.mark.rules
def test_md037_good_valid_emphasis():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/good_valid_emphasis.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_surrounding_emphasis():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_surrounding_emphasis.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_surrounding_emphasis.md:1:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis.md:3:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis.md:5:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis.md:7:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_leading_emphasis():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_leading_emphasis.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_leading_emphasis.md:1:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_leading_emphasis.md:3:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_leading_emphasis.md:5:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_leading_emphasis.md:7:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_trailing_emphasis():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_trailing_emphasis.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_trailing_emphasis.md:1:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_trailing_emphasis.md:3:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_trailing_emphasis.md:5:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_trailing_emphasis.md:7:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_surrounding_emphasis_multiline():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_surrounding_emphasis_multiline.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_surrounding_emphasis_multiline.md:1:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis_multiline.md:4:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis_multiline.md:7:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis_multiline.md:10:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_surrounding_empahsis_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_surrounding_empahsis_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_surrounding_empahsis_setext.md:1:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_empahsis_setext.md:4:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_empahsis_setext.md:7:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_empahsis_setext.md:10:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_surrounding_empahsis_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_surrounding_empahsis_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_surrounding_empahsis_atx.md:1:13: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_empahsis_atx.md:3:14: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_empahsis_atx.md:5:14: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_empahsis_atx.md:7:14: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_bad_surrounding_emphasis_containers():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/bad_surrounding_emphasis_containers.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md037/bad_surrounding_emphasis_containers.md:1:12: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis_containers.md:3:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)\n"
        + "test/resources/rules/md037/bad_surrounding_emphasis_containers.md:5:11: MD037: Spaces inside emphasis markers (no-space-in-emphasis)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_good_emphasis_with_code_span():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/good_emphasis_with_code_span.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md037_good_no_emphasis_but_stars():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md037/good_no_emphasis_but_stars.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to provide tests related to the MD024 rule.
"""
import os
from test.markdown_scanner import MarkdownScanner

import pytest

from .utils import write_temporary_configuration


@pytest.mark.rules
def test_md024_good_different_heading_content_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that all have different
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/different_heading_content_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_content_atx.md:3:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx_with_extra_whitespace():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_atx_with_extra_whitespace.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx_with_extra_emphasis():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_atx_with_extra_emphasis.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx_in_same_list_item():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_atx_in_same_list_item.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_content_atx_in_same_list_item.md:3:3: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx_in_different_list_items():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_atx_in_different_list_items.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_content_atx_in_different_list_items.md:3:3: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx_in_same_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_atx_in_same_block_quote.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_content_atx_in_same_block_quote.md:3:3: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_atx_in_different_block_quotes():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md028",
        "scan",
        "test/resources/rules/md024/same_heading_content_atx_in_different_block_quotes.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_content_atx_in_different_block_quotes.md:3:3: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_in_siblings_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in siblings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_in_siblings_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_in_siblings_atx.md:7:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
        + "test/resources/rules/md024/same_heading_in_siblings_atx.md:11:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_but_not_in_siblings_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text but not in siblings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_but_not_in_siblings_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_but_not_in_siblings_atx.md:9:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_good_different_heading_content_atx_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that all have different
    text in them, with configuration allowing for duplicated text except for in
    sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/different_heading_content_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_good_same_heading_content_atx_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in them, with configuration allowing for duplicated text except for in
    sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_content_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_bad_same_heading_in_siblings_atx_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text in siblings, with configuration allowing for duplicated text except for in
    sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_in_siblings_atx.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md024/same_heading_in_siblings_atx.md:7:1: "
            + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_good_same_heading_but_not_in_siblings_atx_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text but not in siblings, with configuration allowing for duplicated text except for
    in sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_but_not_in_siblings_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_good_same_heading_but_not_in_siblings_atx_with_alternate_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has atx headings that have some duplicated
    text but not in siblings, with the alternate configuration name allowing for
    duplicated text except for in sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"allow_different_nesting": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_but_not_in_siblings_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_good_different_inline_heading_content_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md021 directory that has good atx heading start spacing
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/different_inline_heading_content_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_good_different_heading_content_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that all have different
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/different_heading_content_setext.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_content_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that have some duplicated
    text in them.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md024/same_heading_content_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_content_setext.md:4:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_in_siblings_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that have some duplicated
    text in siblings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md025",
        "scan",
        "test/resources/rules/md024/same_heading_in_siblings_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_in_siblings_setext.md:7:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
        + "test/resources/rules/md024/same_heading_in_siblings_setext.md:13:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_bad_same_heading_but_not_in_siblings_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that have some duplicated
    text but not in siblings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md025",
        "scan",
        "test/resources/rules/md024/same_heading_but_not_in_siblings_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md024/same_heading_but_not_in_siblings_setext.md:10:1: "
        + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md024_good_different_heading_content_setext_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that all have different
    text in them, with configuration allowing for duplicated text except for in
    sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/different_heading_content_setext.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_good_same_heading_content_setext_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that have some
    duplicated text in them, with configuration allowing for duplicated text except for
    in sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_content_setext.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_bad_same_heading_in_siblings_setext_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that have some duplicated
    text in siblings, with configuration allowing for duplicated text except for in
    sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "md025",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_in_siblings_setext.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md024/same_heading_in_siblings_setext.md:7:1: "
            + "MD024: Multiple headings cannot contain the same content. (no-duplicate-heading,no-duplicate-header)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md024_good_same_heading_but_not_in_siblings_setext_with_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD024 directory that has setext headings that have some duplicated
    text but not in siblings, with configuration allowing for duplicated text except for
    in sibling headings.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md024": {"siblings_only": True}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "md025",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md024/same_heading_but_not_in_siblings_setext.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)
